#!/usr/bin/env python
"""
Code for linting modules in the nf-core/modules repository and
in nf-core pipelines

Command:
nf-core modules lint
"""

from __future__ import print_function
import logging
import operator
import os
import questionary
import re
import requests
import rich
import yaml
from rich.console import Console
from rich.table import Table
from rich.markdown import Markdown
import rich
from nf_core.utils import rich_force_colors
from nf_core.lint.pipeline_todos import pipeline_todos
import sys

import nf_core.utils

log = logging.getLogger(__name__)


class ModuleLintException(Exception):
    """Exception raised when there was an error with module linting"""

    pass


class LintResult(object):
    """ An object to hold the results of a lint test """

    def __init__(self, mod, lint_test, message, file_path):
        self.mod = mod
        self.lint_test = lint_test
        self.message = message
        self.file_path = file_path
        self.module_name = mod.module_name


class ModuleLint(object):
    """
    An object for linting modules either in a clone of the 'nf-core/modules'
    repository or in any nf-core pipeline directory
    """

    def __init__(self, dir):
        self.dir = dir
        self.repo_type = self.get_repo_type()
        self.passed = []
        self.warned = []
        self.failed = []

    def lint(self, module=None, all_modules=False, print_results=True, show_passed=False, local=False):
        """
        Lint all or one specific module

        First gets a list of all local modules (in modules/local/process) and all modules
        installed from nf-core (in modules/nf-core/software)

        For all nf-core modules, the correct file structure is assured and important
        file content is verified. If directory subject to linting is a clone of 'nf-core/modules',
        the files necessary for testing the modules are also inspected.

        For all local modules, the '.nf' file is checked for some important flags, and warnings
        are issued if untypical content is found.

        :param module:          A specific module to lint
        :param print_results:   Whether to print the linting results
        :param show_passed:     Whether passed tests should be shown as well

        :returns:               dict of {passed, warned, failed}
        """

        # Get list of all modules in a pipeline
        local_modules, nfcore_modules = self.get_installed_modules()

        # Prompt for module or all
        if module is None and not all_modules:
            question = {
                "type": "list",
                "name": "all_modules",
                "message": "Lint all modules or a single named module?",
                "choices": ["All modules", "Named module"],
            }
            answer = questionary.unsafe_prompt([question], style=nf_core.utils.nfcore_question_style)
            if answer["all_modules"] == "All modules":
                all_modules = True
            else:
                module = questionary.autocomplete(
                    "Tool name:",
                    choices=[m.module_name for m in nfcore_modules],
                    style=nf_core.utils.nfcore_question_style,
                ).ask()

        # Only lint the given module
        if module:
            if all_modules:
                raise ModuleLintException("You cannot specify a tool and request all tools to be linted.")
            local_modules = []
            nfcore_modules = [m for m in nfcore_modules if m.module_name == module]
            if len(nfcore_modules) == 0:
                raise ModuleLintException(f"Could not find the specified module: '{module}'")

        if self.repo_type == "modules":
            log.info(f"Linting modules repo: [magenta]{self.dir}")
        else:
            log.info(f"Linting pipeline: [magenta]{self.dir}")
        if module:
            log.info(f"Linting module: [magenta]{module}")

        # Lint local modules
        if local and len(local_modules) > 0:
            self.lint_local_modules(local_modules)

        # Lint nf-core modules
        if len(nfcore_modules) > 0:
            self.lint_nfcore_modules(nfcore_modules)

            self.check_module_changes(nfcore_modules)

        if print_results:
            self._print_results(show_passed=show_passed)

        return {"passed": self.passed, "warned": self.warned, "failed": self.failed}

    def lint_local_modules(self, local_modules):
        """
        Lint a local module
        Only issues warnings instead of failures
        """
        progress_bar = rich.progress.Progress(
            "[bold blue]{task.description}",
            rich.progress.BarColumn(bar_width=None),
            "[magenta]{task.completed} of {task.total}[reset] » [bold yellow]{task.fields[test_name]}",
            transient=True,
        )
        with progress_bar:
            lint_progress = progress_bar.add_task(
                "Linting local modules", total=len(local_modules), test_name=os.path.basename(local_modules[0])
            )

            for mod in local_modules:
                progress_bar.update(lint_progress, advance=1, test_name=os.path.basename(mod))
                mod_object = NFCoreModule(
                    module_dir=mod, base_dir=self.dir, repo_type=self.repo_type, nf_core_module=False
                )
                mod_object.main_nf = mod
                mod_object.module_name = os.path.basename(mod)
                mod_object.lint_main_nf()
                self.passed += [LintResult(mod_object, m[0], m[1], m[2]) for m in mod_object.passed]
                self.warned += [LintResult(mod_object, m[0], m[1], m[2]) for m in mod_object.warned + mod_object.failed]

    def lint_nfcore_modules(self, nfcore_modules):
        """
        Lint nf-core modules
        For each nf-core module, checks for existence of the files
        - main.nf
        - meta.yml
        - functions.nf
        And verifies that their content.

        If the linting is run for modules in the central nf-core/modules repo
        (repo_type==modules), files that are relevant for module testing are
        also examined
        """

        progress_bar = rich.progress.Progress(
            "[bold blue]{task.description}",
            rich.progress.BarColumn(bar_width=None),
            "[magenta]{task.completed} of {task.total}[reset] » [bold yellow]{task.fields[test_name]}",
            transient=True,
        )
        with progress_bar:
            lint_progress = progress_bar.add_task(
                "Linting nf-core modules", total=len(nfcore_modules), test_name=nfcore_modules[0].module_name
            )
            for mod in nfcore_modules:
                progress_bar.update(lint_progress, advance=1, test_name=mod.module_name)
                passed, warned, failed = mod.lint()
                self.passed += [LintResult(mod, m[0], m[1], m[2]) for m in passed]
                self.warned += [LintResult(mod, m[0], m[1], m[2]) for m in warned]
                self.failed += [LintResult(mod, m[0], m[1], m[2]) for m in failed]

    def get_repo_type(self):
        """
        Determine whether this is a pipeline repository or a clone of
        nf-core/modules
        """
        # Verify that the pipeline dir exists
        if self.dir is None or not os.path.exists(self.dir):
            log.error("Could not find directory: {}".format(self.dir))
            sys.exit(1)

        # Determine repository type
        if os.path.exists(os.path.join(self.dir, "main.nf")):
            return "pipeline"
        elif os.path.exists(os.path.join(self.dir, "software")):
            return "modules"
        else:
            log.error("Could not determine repository type of {}".format(self.dir))
            sys.exit(1)

    def get_installed_modules(self):
        """
        Make a list of all modules installed in this repository

        Returns a tuple of two lists, one for local modules
        and one for nf-core modules. The local modules are represented
        as direct filepaths to the module '.nf' file.
        Nf-core module are returned as file paths to the module directories.
        In case the module contains several tools, one path to each tool directory
        is returned.

        returns (local_modules, nfcore_modules)
        """
        # initialize lists
        local_modules = []
        nfcore_modules = []
        local_modules_dir = None
        nfcore_modules_dir = os.path.join(self.dir, "modules", "nf-core", "software")

        # Get local modules
        if self.repo_type == "pipeline":
            local_modules_dir = os.path.join(self.dir, "modules", "local", "process")

            # Filter local modules
            if os.path.exists(local_modules_dir):
                local_modules = os.listdir(local_modules_dir)
                local_modules = sorted([x for x in local_modules if (x.endswith(".nf") and not x == "functions.nf")])

        # nf-core/modules
        if self.repo_type == "modules":
            nfcore_modules_dir = os.path.join(self.dir, "software")

        # Get nf-core modules
        if os.path.exists(nfcore_modules_dir):
            for m in sorted([m for m in os.listdir(nfcore_modules_dir) if not m == "lib"]):
                if not os.path.isdir(os.path.join(nfcore_modules_dir, m)):
                    raise ModuleLintException(
                        f"File found in '{nfcore_modules_dir}': '{m}'! This directory should only contain module directories."
                    )
                m_content = os.listdir(os.path.join(nfcore_modules_dir, m))
                # Not a module, but contains sub-modules
                if not "main.nf" in m_content:
                    for tool in m_content:
                        nfcore_modules.append(os.path.join(m, tool))
                else:
                    nfcore_modules.append(m)

        # Make full (relative) file paths and create NFCoreModule objects
        local_modules = [os.path.join(local_modules_dir, m) for m in local_modules]
        nfcore_modules = [
            NFCoreModule(os.path.join(nfcore_modules_dir, m), repo_type=self.repo_type, base_dir=self.dir)
            for m in nfcore_modules
        ]

        return local_modules, nfcore_modules

    def _print_results(self, show_passed=False):
        """Print linting results to the command line.

        Uses the ``rich`` library to print a set of formatted tables to the command line
        summarising the linting results.
        """

        log.debug("Printing final results")
        console = Console(force_terminal=rich_force_colors())

        # Sort the results
        self.passed.sort(key=operator.attrgetter("message", "module_name"))
        self.warned.sort(key=operator.attrgetter("message", "module_name"))
        self.failed.sort(key=operator.attrgetter("message", "module_name"))

        # Find maximum module name length
        max_mod_name_len = 40
        for idx, tests in enumerate([self.passed, self.warned, self.failed]):
            try:
                for lint_result in tests:
                    max_mod_name_len = max(len(lint_result.module_name), max_mod_name_len)
            except:
                pass

        # Helper function to format test links nicely
        def format_result(test_results, table):
            """
            Given an list of error message IDs and the message texts, return a nicely formatted
            string for the terminal with appropriate ASCII colours.
            """
            # TODO: Row styles don't work current as table-level style overrides.
            # I'd like to make an issue about this on the rich repo so leaving here in case there is a future fix
            last_modname = False
            row_style = None
            for lint_result in test_results:
                if last_modname and lint_result.module_name != last_modname:
                    if row_style:
                        row_style = None
                    else:
                        row_style = "magenta"
                last_modname = lint_result.module_name
                table.add_row(
                    Markdown(f"{lint_result.module_name}"),
                    os.path.relpath(lint_result.file_path, self.dir),
                    Markdown(f"{lint_result.message}"),
                    style=row_style,
                )
            return table

        def _s(some_list):
            if len(some_list) > 1:
                return "s"
            return ""

        # Table of passed tests
        if len(self.passed) > 0 and show_passed:
            console.print(
                rich.panel.Panel(r"[!] {} Test{} Passed".format(len(self.passed), _s(self.passed)), style="bold green")
            )
            table = Table(style="green", box=rich.box.ROUNDED)
            table.add_column("Module name", width=max_mod_name_len)
            table.add_column("File path")
            table.add_column("Test message")
            table = format_result(self.passed, table)
            console.print(table)

        # Table of warning tests
        if len(self.warned) > 0:
            console.print(
                rich.panel.Panel(
                    r"[!] {} Test Warning{}".format(len(self.warned), _s(self.warned)), style="bold yellow"
                )
            )
            table = Table(style="yellow", box=rich.box.ROUNDED)
            table.add_column("Module name", width=max_mod_name_len)
            table.add_column("File path")
            table.add_column("Test message")
            table = format_result(self.warned, table)
            console.print(table)

        # Table of failing tests
        if len(self.failed) > 0:
            console.print(
                rich.panel.Panel(r"[!] {} Test{} Failed".format(len(self.failed), _s(self.failed)), style="bold red")
            )
            table = Table(style="red", box=rich.box.ROUNDED)
            table.add_column("Module name", width=max_mod_name_len)
            table.add_column("File path")
            table.add_column("Test message")
            table = format_result(self.failed, table)
            console.print(table)

        # Summary table
        table = Table(box=rich.box.ROUNDED)
        table.add_column("[bold green]LINT RESULTS SUMMARY".format(len(self.passed)), no_wrap=True)
        table.add_row(
            r"[✔] {:>3} Test{} Passed".format(len(self.passed), _s(self.passed)),
            style="green",
        )
        table.add_row(r"[!] {:>3} Test Warning{}".format(len(self.warned), _s(self.warned)), style="yellow")
        table.add_row(r"[✗] {:>3} Test{} Failed".format(len(self.failed), _s(self.failed)), style="red")
        console.print(table)

    def check_module_changes(self, nfcore_modules):
        """
        Checks whether installed nf-core modules have changed compared to the
        original repository
        Downloads the 'main.nf', 'functions.nf' and 'meta.yml' files for every module
        and compare them to the local copies
        """
        files_to_check = ["main.nf", "functions.nf", "meta.yml"]

        progress_bar = rich.progress.Progress(
            "[bold blue]{task.description}",
            rich.progress.BarColumn(bar_width=None),
            "[magenta]{task.completed} of {task.total}[reset] » [bold yellow]{task.fields[test_name]}",
            transient=True,
        )
        with progress_bar:
            comparison_progress = progress_bar.add_task(
                "Comparing local file to remote", total=len(nfcore_modules), test_name=nfcore_modules[0].module_name
            )
            # Loop over nf-core modules
            for mod in nfcore_modules:
                progress_bar.update(comparison_progress, advance=1, test_name=mod.module_name)
                module_base_url = (
                    f"https://raw.githubusercontent.com/nf-core/modules/master/software/{mod.module_name}/"
                )

                for f in files_to_check:
                    # open local copy, continue if file not found (a failed message has already been issued in this case)
                    try:
                        local_copy = open(os.path.join(mod.module_dir, f), "r").read()
                    except FileNotFoundError as e:
                        continue

                    # Download remote copy and compare
                    url = module_base_url + f
                    r = requests.get(url=url)

                    if r.status_code != 200:
                        self.warned.append(
                            LintResult(
                                mod,
                                "check_local_copy",
                                f"Could not fetch remote copy, skipping comparison.",
                                f"{os.path.join(mod.module_dir, f)}",
                            )
                        )
                    else:
                        try:
                            remote_copy = r.content.decode("utf-8")

                            if local_copy != remote_copy:
                                self.warned.append(
                                    LintResult(
                                        mod,
                                        "check_local_copy",
                                        "Local copy of module outdated",
                                        f"{os.path.join(mod.module_dir, f)}",
                                    )
                                )
                            else:
                                self.passed.append(
                                    LintResult(
                                        mod,
                                        "check_local_copy",
                                        "Local copy of module up to date",
                                        f"{os.path.join(mod.module_dir, f)}",
                                    )
                                )
                        except UnicodeDecodeError as e:
                            self.warned.append(
                                LintResult(
                                    mod,
                                    "check_local_copy",
                                    f"Could not decode file from {url}. Skipping comparison ({e})",
                                    f"{os.path.join(mod.module_dir, f)}",
                                )
                            )


class NFCoreModule(object):
    """
    A class to hold the information a bout a nf-core module
    Includes functionality for linting
    """

    def __init__(self, module_dir, repo_type, base_dir, nf_core_module=True):
        self.module_dir = module_dir
        self.repo_type = repo_type
        self.base_dir = base_dir
        self.passed = []
        self.warned = []
        self.failed = []
        self.inputs = []
        self.outputs = []
        self.has_meta = False

        if nf_core_module:
            # Initialize the important files
            self.main_nf = os.path.join(self.module_dir, "main.nf")
            self.meta_yml = os.path.join(self.module_dir, "meta.yml")
            self.function_nf = os.path.join(self.module_dir, "functions.nf")
            self.software = self.module_dir.split("software" + os.sep)[1]
            self.test_dir = os.path.join(self.base_dir, "tests", "software", self.software)
            self.test_yml = os.path.join(self.test_dir, "test.yml")
            self.test_main_nf = os.path.join(self.test_dir, "main.nf")
            self.module_name = module_dir.split("software" + os.sep)[1]

    def lint(self):
        """ Perform linting on this module """
        # Iterate over modules and run all checks on them

        # Lint the main.nf file
        self.lint_main_nf()

        # Lint the meta.yml file
        self.lint_meta_yml()

        # Lint the functions.nf file
        self.lint_functions_nf()

        # Lint the tests
        if self.repo_type == "modules":
            self.lint_module_tests()

        # Check for TODOs
        self.wf_path = self.module_dir
        module_todos = pipeline_todos(self)
        for i, warning in enumerate(module_todos["warned"]):
            self.warned.append(("module_todo", warning, module_todos["file_paths"][i]))

        return self.passed, self.warned, self.failed

    def lint_module_tests(self):
        """ Lint module tests """

        if os.path.exists(self.test_dir):
            self.passed.append(("test_dir_exists", "Test directory exists", self.test_dir))
        else:
            self.failed.append(("test_dir_exists", "Test directory is missing", self.test_dir))
            return

        # Lint the test main.nf file
        test_main_nf = os.path.join(self.test_dir, "main.nf")
        if os.path.exists(test_main_nf):
            self.passed.append(("test_main_exists", "test `main.nf` exists", self.test_main_nf))
        else:
            self.failed.append(("test_main_exists", "test `main.nf` does not exist", self.test_main_nf))

        # Lint the test.yml file
        try:
            with open(self.test_yml, "r") as fh:
                test_yml = yaml.safe_load(fh)
            self.passed.append(("test_yml_exists", "Test `test.yml` exists", self.test_yml))
        except FileNotFoundError:
            self.failed.append(("test_yml_exists", "Test `test.yml` does not exist", self.test_yml))

    def lint_meta_yml(self):
        """ Lint a meta yml file """
        required_keys = ["name", "input", "output"]
        required_keys_lists = ["intput", "output"]
        try:
            with open(self.meta_yml, "r") as fh:
                meta_yaml = yaml.safe_load(fh)
            self.passed.append(("meta_yml_exists", "Module `meta.yml` exists", self.meta_yml))
        except FileNotFoundError:
            self.failed.append(("meta_yml_exists", "Module `meta.yml` does not exist", self.meta_yml))
            return

        # Confirm that all required keys are given
        contains_required_keys = True
        all_list_children = True
        for rk in required_keys:
            if not rk in meta_yaml.keys():
                self.failed.append(("meta_required_keys", f"`{rk}` not specified", self.meta_yml))
                contains_required_keys = False
            elif not isinstance(meta_yaml[rk], list) and rk in required_keys_lists:
                self.failed.append(("meta_required_keys", f"`{rk}` is not a list", self.meta_yml))
                all_list_children = False
        if contains_required_keys:
            self.passed.append(("meta_required_keys", "`meta.yml` contains all required keys", self.meta_yml))

        # Confirm that all input and output channels are specified
        if contains_required_keys and all_list_children:
            meta_input = [list(x.keys())[0] for x in meta_yaml["input"]]
            for input in self.inputs:
                if input in meta_input:
                    self.passed.append(("meta_input", f"`{input}` specified", self.meta_yml))
                else:
                    self.failed.append(("meta_input", f"`{input}` missing in `meta.yml`", self.meta_yml))

            meta_output = [list(x.keys())[0] for x in meta_yaml["output"]]
            for output in self.outputs:
                if output in meta_output:
                    self.passed.append(("meta_output", "`{output}` specified", self.meta_yml))
                else:
                    self.failed.append(("meta_output", "`{output}` missing in `meta.yml`", self.meta_yml))

            # confirm that the name matches the process name in main.nf
            if meta_yaml["name"].upper() == self.process_name:
                self.passed.append(("meta_name", "Correct name specified in `meta.yml`", self.meta_yml))
            else:
                self.failed.append(
                    ("meta_name", "Conflicting process name between `meta.yml` and `main.nf`", self.meta_yml)
                )

    def lint_main_nf(self):
        """
        Lint a single main.nf module file
        Can also be used to lint local module files,
        in which case failures should be interpreted
        as warnings
        """
        inputs = []
        outputs = []

        # Check whether file exists and load it
        try:
            with open(self.main_nf, "r") as fh:
                lines = fh.readlines()
            self.passed.append(("main_nf_exists", "Module file exists", self.main_nf))
        except FileNotFoundError as e:
            self.failed.append(("main_nf_exists", "Module file does not exist", self.main_nf))
            return

        # Check that options are defined
        initoptions_re = re.compile(r"\s*options\s*=\s*initOptions\s*\(\s*params\.options\s*\)\s*")
        paramsoptions_re = re.compile(r"\s*params\.options\s*=\s*\[:\]\s*")
        if any(initoptions_re.match(l) for l in lines) and any(paramsoptions_re.match(l) for l in lines):
            self.passed.append(("main_nf_options", "'options' variable specified", self.main_nf))
        else:
            self.warned.append(("main_nf_options", "'options' variable not specified", self.main_nf))

        # Go through module main.nf file and switch state according to current section
        # Perform section-specific linting
        state = "module"
        process_lines = []
        script_lines = []
        for l in lines:
            if re.search("^\s*process\s*\w*\s*{", l) and state == "module":
                state = "process"
            if re.search("input\s*:", l) and state == "process":
                state = "input"
                continue
            if re.search("output\s*:", l) and state == "input":
                state = "output"
                continue
            if re.search("script\s*:", l) and state == "output":
                state = "script"
                continue

            # Perform state-specific linting checks
            if state == "process" and not self._is_empty(l):
                process_lines.append(l)
            if state == "input" and not self._is_empty(l):
                inputs += self._parse_input(l)
            if state == "output" and not self._is_empty(l):
                outputs += self._parse_output(l)
                outputs = list(set(outputs))  # remove duplicate 'meta's
            if state == "script" and not self._is_empty(l):
                script_lines.append(l)

        # Check the process definitions
        if self.check_process_section(process_lines):
            self.passed.append(("main_nf_container", "Container versions match", self.main_nf))
        else:
            self.warned.append(("main_nf_container", "Container versions do not match", self.main_nf))

        # Check the script definition
        self.check_script_section(script_lines)

        # Check whether 'meta' is emitted when given as input
        if "meta" in inputs:
            self.has_meta = True
            if "meta" in outputs:
                self.passed.append(("main_nf_meta_output", "'meta' map emitted in output channel(s)", self.main_nf))
            else:
                self.failed.append(("main_nf_meta_output", "'meta' map not emitted in output channel(s)", self.main_nf))

            # if meta is specified, it should also be used as 'saveAs ... publishId:meta.id'
            save_as = [pl for pl in process_lines if "saveAs" in pl]
            if len(save_as) > 0 and re.search("\s*publish_id\s*:\s*meta.id", save_as[0]):
                self.passed.append(("main_nf_meta_saveas", "'meta.id' specified in saveAs function", self.main_nf))
            else:
                self.failed.append(("main_nf_meta_saveas", "'meta.id' unspecificed in saveAs function", self.main_nf))

        # Check that a software version is emitted
        if "version" in outputs:
            self.passed.append(("main_nf_version_emitted", "Module emits software version", self.main_nf))
        else:
            self.warned.append(("main_nf_version_emitted", "Module does not emit software version", self.main_nf))

        return inputs, outputs

    def check_script_section(self, lines):
        """
        Lint the script section
        Checks whether 'def sotware' and 'def prefix' are defined
        """
        script = "".join(lines)

        # check for software
        if re.search("\s*def\s*software\s*=\s*getSoftwareName", script):
            self.passed.append(("main_nf_version_script", "Software version specified in script section", self.main_nf))
        else:
            self.warned.append(
                ("main_nf_version_script", "Software version unspecified in script section", self.main_nf)
            )

        # check for prefix (only if module has a meta map as input)
        if self.has_meta:
            if re.search("\s*prefix\s*=\s*options.suffix", script):
                self.passed.append(("main_nf_meta_prefix", "'prefix' specified in script section", self.main_nf))
            else:
                self.failed.append(("main_nf_meta_prefix", "'prefix' unspecified in script section", self.main_nf))

    def check_process_section(self, lines):
        """
        Lint the section of a module between the process definition
        and the 'input:' definition
        Specifically checks for correct software versions
        and containers
        """
        # Checks that build numbers of bioconda, singularity and docker container are matching
        build_id = "build"
        singularity_tag = "singularity"
        docker_tag = "docker"
        bioconda_packages = []

        # Process name should be all capital letters
        self.process_name = lines[0].split()[1]
        if all([x.upper() for x in self.process_name]):
            self.passed.append(("process_capitals", "Process name is in capital letters", self.main_nf))
        else:
            self.failed.append(("process_capitals", "Process name is not in captial letters", self.main_nf))

        # Check that process labels are correct
        correct_process_labels = ["process_low", "process_medium", "process_high", "process_long"]
        process_label = [l for l in lines if "label" in l]
        if len(process_label) > 0:
            process_label = process_label[0].split()[1].strip().strip("'").strip('"')
            if not process_label in correct_process_labels:
                self.warned.append(
                    (
                        "process_standard_label",
                        f"Process label ({process_label}) is not among standard labels: `{'`,`'.join(correct_process_labels)}`",
                        self.main_nf,
                    )
                )
            else:
                self.passed.append(("process_standard_label", "Correct process label", self.main_nf))
        else:
            self.warned.append(("process_standard_label", "Process label unspecified", self.main_nf))

        for l in lines:
            if re.search("bioconda::", l):
                bioconda_packages = [b for b in l.split() if "bioconda::" in b]
            if re.search("org/singularity", l):
                singularity_tag = l.split("/")[-1].replace('"', "").replace("'", "").split("--")[-1].strip()
            if re.search("biocontainers", l):
                docker_tag = l.split("/")[-1].replace('"', "").replace("'", "").split("--")[-1].strip()

        # Check that all bioconda packages have build numbers
        # Also check for newer versions
        for bp in bioconda_packages:
            bp = bp.strip("'").strip('"')
            # Check for correct version and newer versions
            try:
                bioconda_version = bp.split("=")[1]
                # response = _bioconda_package(bp)
                response = nf_core.utils.anaconda_package(bp)
            except LookupError as e:
                self.warned.append(("bioconda_version", "Conda version not specified correctly", self.main_nf))
            except ValueError as e:
                self.failed.append(("bioconda_version", "Conda version not specified correctly", self.main_nf))
            else:
                # Check that required version is available at all
                if bioconda_version not in response.get("versions"):
                    self.failed.append(("bioconda_version", "Conda package had unknown version: `{}`", self.main_nf))
                    continue  # No need to test for latest version, continue linting
                # Check version is latest available
                last_ver = response.get("latest_version")
                if last_ver is not None and last_ver != bioconda_version:
                    package, ver = bp.split("=", 1)
                    self.warned.append(
                        ("bioconda_latest", f"Conda update: {package} `{ver}` -> `{last_ver}`", self.main_nf)
                    )
                else:
                    self.passed.append(
                        ("bioconda_latest", "Conda package is the latest available: `{bp}`", self.main_nf)
                    )

        if docker_tag == singularity_tag:
            return True
        else:
            return False

    def lint_functions_nf(self):
        """
        Lint a functions.nf file
        Verifies that the file exists and contains all necessary functions
        """
        try:
            with open(self.function_nf, "r") as fh:
                lines = fh.readlines()
            self.passed.append(("functions_nf_exists", "'functions.nf' exists", self.function_nf))
        except FileNotFoundError as e:
            self.failed.append(("functions_nf_exists", "'functions.nf' does not exist", self.function_nf))
            return

        # Test whether all required functions are present
        required_functions = ["getSoftwareName", "initOptions", "getPathFromList", "saveFiles"]
        lines = "\n".join(lines)
        contains_all_functions = True
        for f in required_functions:
            if not "def " + f in lines:
                self.failed.append(("functions_nf_func_exist", "Function is missing: `{f}`", self.function_nf))
                contains_all_functions = False
        if contains_all_functions:
            self.passed.append(("functions_nf_func_exist", "All functions present", self.function_nf))

    def _parse_input(self, line):
        input = []
        # more than one input
        if "tuple" in line:
            line = line.replace("tuple", "")
            line = line.replace(" ", "")
            line = line.split(",")

            for elem in line:
                elem = elem.split("(")[1]
                elem = elem.replace(")", "").strip()
                input.append(elem)
        else:
            if "(" in line:
                input.append(line.split("(")[1].replace(")", ""))
            else:
                input.append(line.split()[1])
        return input

    def _parse_output(self, line):
        output = []
        if "meta" in line:
            output.append("meta")
        if not "emit" in line:
            self.failed.append(("missing_emit", f"Missing emit statement: {line.strip()}", self.main_nf))
        if "emit" in line:
            output.append(line.split("emit:")[1].strip())

        return output

    def _is_empty(self, line):
        """ Check whether a line is empty or a comment """
        empty = False
        if line.strip().startswith("//"):
            empty = True
        if line.strip().replace(" ", "") == "":
            empty = True
        return empty



#!/usr/bin/env python
from __future__ import print_function
from collections import OrderedDict
import re

# TODO nf-core: Add additional regexes for new tools in process get_software_versions
regexes = {
    "{{ name }}": ["v_pipeline.txt", r"(\S+)"],
    "Nextflow": ["v_nextflow.txt", r"(\S+)"],
    "FastQC": ["v_fastqc.txt", r"FastQC v(\S+)"],
    "MultiQC": ["v_multiqc.txt", r"multiqc, version (\S+)"],
}
results = OrderedDict()
results["{{ name }}"] = '<span style="color:#999999;">N/A</span>'
results["Nextflow"] = '<span style="color:#999999;">N/A</span>'
results["FastQC"] = '<span style="color:#999999;">N/A</span>'
results["MultiQC"] = '<span style="color:#999999;">N/A</span>'

# Search each file using its regex
for k, v in regexes.items():
    try:
        with open(v[0]) as x:
            versions = x.read()
            match = re.search(v[1], versions)
            if match:
                results[k] = "v{}".format(match.group(1))
    except IOError:
        results[k] = False

# Remove software set to false in results
for k in list(results):
    if not results[k]:
        del results[k]

# Dump to YAML
print(
    """
id: 'software_versions'
section_name: '{{ name }} Software Versions'
section_href: 'https://github.com/{{ name }}'
plot_type: 'html'
description: 'are collected at run time from the software output.'
data: |
    <dl class="dl-horizontal">
"""
)
for k, v in results.items():
    print("        <dt>{}</dt><dd><samp>{}</samp></dd>".format(k, v))
print("    </dl>")

# Write out regexes as csv file:
with open("software_versions.csv", "w") as f:
    for k, v in results.items():
        f.write("{}\t{}\n".format(k, v))



#!/usr/bin/env python

import os
import yaml


def actions_awsfulltest(self):
    """Checks the GitHub Actions awsfulltest is valid.

    In addition to small test datasets run on GitHub Actions, we provide the possibility of testing the pipeline on full size datasets on AWS.
    This should ensure that the pipeline runs as expected on AWS and provide a resource estimation.

    The GitHub Actions workflow is called ``awsfulltest.yml``, and it can be found in the ``.github/workflows/`` directory.

    .. warning::  This workflow incurs AWS costs, therefore it should only be triggered for pipeline releases:
                  ``workflow_run`` (after the docker hub release workflow) and ``workflow_dispatch``.

    .. note::  You can manually trigger the AWS tests by going to the `Actions` tab on the pipeline GitHub repository and selecting the
                  `nf-core AWS full size tests` workflow on the left.

    .. tip::  For tests on full data prior to release, `Nextflow Tower <https://tower.nf>`_ launch feature can be employed.

    The ``.github/workflows/awsfulltest.yml`` file is tested for the following:

    * Must be turned on ``workflow_dispatch``.
    * Must be turned on for ``workflow_run`` with ``workflows: ["nf-core Docker push (release)"]`` and ``types: [completed]``
    * Should run the profile ``test_full`` that should be edited to provide the links to full-size datasets. If it runs the profile ``test``, a warning is given.
    """
    passed = []
    warned = []
    failed = []

    fn = os.path.join(self.wf_path, ".github", "workflows", "awsfulltest.yml")
    if os.path.isfile(fn):
        try:
            with open(fn, "r") as fh:
                wf = yaml.safe_load(fh)
        except Exception as e:
            return {"failed": ["Could not parse yaml file: {}, {}".format(fn, e)]}

        aws_profile = "-profile test "

        # Check that the action is only turned on for published releases
        try:
            assert "workflow_run" in wf[True]
            assert wf[True]["workflow_run"]["workflows"] == ["nf-core Docker push (release)"]
            assert wf[True]["workflow_run"]["types"] == ["completed"]
            assert "workflow_dispatch" in wf[True]
        except (AssertionError, KeyError, TypeError):
            failed.append("`.github/workflows/awsfulltest.yml` is not triggered correctly")
        else:
            passed.append("`.github/workflows/awsfulltest.yml` is triggered correctly")

        # Warn if `-profile test` is still unchanged
        try:
            steps = wf["jobs"]["run-awstest"]["steps"]
            assert any([aws_profile in step["run"] for step in steps if "run" in step.keys()])
        except (AssertionError, KeyError, TypeError):
            passed.append("`.github/workflows/awsfulltest.yml` does not use `-profile test`")
        else:
            warned.append("`.github/workflows/awsfulltest.yml` should test full datasets, not `-profile test`")

    return {"passed": passed, "warned": warned, "failed": failed}



#!/usr/bin/env python

import os
import re
import yaml


def actions_ci(self):
    """Checks that the GitHub Actions pipeline CI (Continuous Integration) workflow is valid.

    The ``.github/workflows/ci.yml`` GitHub Actions workflow runs the pipeline on a minimal test
    dataset using ``-profile test`` to check that no breaking changes have been introduced.
    Final result files are not checked, just that the pipeline exists successfully.

    This lint test checks this GitHub Actions workflow file for the following:

    * Workflow must be triggered on the following events:

      .. code-block:: yaml

         on:
             push:
             branches:
                 - dev
             pull_request:
             release:
             types: [published]

    * The minimum Nextflow version specified in the pipeline's ``nextflow.config`` matches that defined by ``nxf_ver`` in the test matrix:

      .. code-block:: yaml
         :emphasize-lines: 4

         strategy:
           matrix:
             # Nextflow versions: check pipeline minimum and current latest
             nxf_ver: ['19.10.0', '']

      .. note:: These ``matrix`` variables run the test workflow twice, varying the ``nxf_ver`` variable each time.
                This is used in the ``nextflow run`` commands to test the pipeline with both the latest available version
                of the pipeline (``''``) and the stated minimum required version.

    * The `Docker` container for the pipeline must use the correct pipeline version number:

        * Development pipelines:

          .. code-block:: bash

             docker tag nfcore/<pipeline_name>:dev nfcore/<pipeline_name>:dev

        * Released pipelines:

          .. code-block:: bash

             docker tag nfcore/<pipeline_name>:dev nfcore/<pipeline_name>:<pipeline-version>

        * Complete example for a released pipeline called *nf-core/example* with version number ``1.0.0``:

          .. code-block:: yaml
             :emphasize-lines: 3,8,9

             - name: Build new docker image
               if: env.GIT_DIFF
               run: docker build --no-cache . -t nfcore/example:1.0.0

             - name: Pull docker image
               if: ${{ !env.GIT_DIFF }}
               run: |
                 docker pull nfcore/example:dev
                 docker tag nfcore/example:dev nfcore/example:1.0.0
    """
    passed = []
    failed = []
    fn = os.path.join(self.wf_path, ".github", "workflows", "ci.yml")

    # Return an ignored status if we can't find the file
    if not os.path.isfile(fn):
        return {"ignored": ["'.github/workflows/ci.yml' not found"]}

    try:
        with open(fn, "r") as fh:
            ciwf = yaml.safe_load(fh)
    except Exception as e:
        return {"failed": ["Could not parse yaml file: {}, {}".format(fn, e)]}

    # Check that the action is turned on for the correct events
    try:
        expected = {"push": {"branches": ["dev"]}, "pull_request": None, "release": {"types": ["published"]}}
        # NB: YAML dict key 'on' is evaluated to a Python dict key True
        assert ciwf[True] == expected
    except (AssertionError, KeyError, TypeError):
        failed.append("'.github/workflows/ci.yml' is not triggered on expected events")
    else:
        passed.append("'.github/workflows/ci.yml' is triggered on expected events")

    # Check that we're pulling the right docker image and tagging it properly
    if self.nf_config.get("process.container", ""):
        docker_notag = re.sub(r":(?:[\.\d]+|dev)$", "", self.nf_config.get("process.container", "").strip("\"'"))
        docker_withtag = self.nf_config.get("process.container", "").strip("\"'")

        # docker build
        docker_build_cmd = "docker build --no-cache . -t {}".format(docker_withtag)
        try:
            steps = ciwf["jobs"]["test"]["steps"]
            assert any([docker_build_cmd in step["run"] for step in steps if "run" in step.keys()])
        except (AssertionError, KeyError, TypeError):
            failed.append("CI is not building the correct docker image. Should be: `{}`".format(docker_build_cmd))
        else:
            passed.append("CI is building the correct docker image: `{}`".format(docker_build_cmd))

        # docker pull
        docker_pull_cmd = "docker pull {}:dev".format(docker_notag)
        try:
            steps = ciwf["jobs"]["test"]["steps"]
            assert any([docker_pull_cmd in step["run"] for step in steps if "run" in step.keys()])
        except (AssertionError, KeyError, TypeError):
            failed.append("CI is not pulling the correct docker image. Should be: `{}`".format(docker_pull_cmd))
        else:
            passed.append("CI is pulling the correct docker image: {}".format(docker_pull_cmd))

        # docker tag
        docker_tag_cmd = "docker tag {}:dev {}".format(docker_notag, docker_withtag)
        try:
            steps = ciwf["jobs"]["test"]["steps"]
            assert any([docker_tag_cmd in step["run"] for step in steps if "run" in step.keys()])
        except (AssertionError, KeyError, TypeError):
            failed.append("CI is not tagging docker image correctly. Should be: `{}`".format(docker_tag_cmd))
        else:
            passed.append("CI is tagging docker image correctly: {}".format(docker_tag_cmd))

    # Check that we are testing the minimum nextflow version
    try:
        matrix = ciwf["jobs"]["test"]["strategy"]["matrix"]["nxf_ver"]
        assert any([self.minNextflowVersion in matrix])
    except (KeyError, TypeError):
        failed.append("'.github/workflows/ci.yml' does not check minimum NF version")
    except AssertionError:
        failed.append("Minimum NF version in '.github/workflows/ci.yml' different to pipeline's manifest")
    else:
        passed.append("'.github/workflows/ci.yml' checks minimum NF version")

    return {"passed": passed, "failed": failed}



#!/usr/bin/env python
"""
Code to handle several functions in order to deal with nf-core/modules in
nf-core pipelines

* list modules
* install modules
* remove modules
* update modules (TODO)
*
"""

from __future__ import print_function
import base64
import glob
import json
import logging
import os
import re
import hashlib
import questionary
import requests
import rich
import shutil
import yaml
from rich.console import Console
from rich.table import Table
from rich.markdown import Markdown
import rich
from nf_core.utils import rich_force_colors
from nf_core.lint.pipeline_todos import pipeline_todos
import sys

import nf_core.utils

log = logging.getLogger(__name__)


class ModulesRepo(object):
    """
    An object to store details about the repository being used for modules.

    Used by the `nf-core modules` top-level command with -r and -b flags,
    so that this can be used in the same way by all sucommands.
    """

    def __init__(self, repo="nf-core/modules", branch="master"):
        self.name = repo
        self.branch = branch
        self.modules_file_tree = {}
        self.modules_current_hash = None
        self.modules_avail_module_names = []

    def get_modules_file_tree(self):
        """
        Fetch the file list from the repo, using the GitHub API

        Sets self.modules_file_tree
             self.modules_current_hash
             self.modules_avail_module_names
        """
        api_url = "https://api.github.com/repos/{}/git/trees/{}?recursive=1".format(self.name, self.branch)
        r = requests.get(api_url, auth=nf_core.utils.github_api_auto_auth())
        if r.status_code == 404:
            log.error("Repository / branch not found: {} ({})\n{}".format(self.name, self.branch, api_url))
            sys.exit(1)
        elif r.status_code != 200:
            raise SystemError(
                "Could not fetch {} ({}) tree: {}\n{}".format(self.name, self.branch, r.status_code, api_url)
            )

        result = r.json()
        assert result["truncated"] == False

        self.modules_current_hash = result["sha"]
        self.modules_file_tree = result["tree"]
        for f in result["tree"]:
            if f["path"].startswith("software/") and f["path"].endswith("/main.nf") and "/test/" not in f["path"]:
                # remove software/ and /main.nf
                self.modules_avail_module_names.append(f["path"][9:-8])

    def get_module_file_urls(self, module):
        """Fetch list of URLs for a specific module

        Takes the name of a module and iterates over the GitHub repo file tree.
        Loops over items that are prefixed with the path 'software/<module_name>' and ignores
        anything that's not a blob. Also ignores the test/ subfolder.

        Returns a dictionary with keys as filenames and values as GitHub API URIs.
        These can be used to then download file contents.

        Args:
            module (string): Name of module for which to fetch a set of URLs

        Returns:
            dict: Set of files and associated URLs as follows:

            {
                'software/fastqc/main.nf': 'https://api.github.com/repos/nf-core/modules/git/blobs/65ba598119206a2b851b86a9b5880b5476e263c3',
                'software/fastqc/meta.yml': 'https://api.github.com/repos/nf-core/modules/git/blobs/0d5afc23ba44d44a805c35902febc0a382b17651'
            }
        """
        results = {}
        for f in self.modules_file_tree:
            if not f["path"].startswith("software/{}".format(module)):
                continue
            if f["type"] != "blob":
                continue
            if "/test/" in f["path"]:
                continue
            results[f["path"]] = f["url"]
        return results

    def download_gh_file(self, dl_filename, api_url):
        """Download a file from GitHub using the GitHub API

        Args:
            dl_filename (string): Path to save file to
            api_url (string): GitHub API URL for file

        Raises:
            If a problem, raises an error
        """

        # Make target directory if it doesn't already exist
        dl_directory = os.path.dirname(dl_filename)
        if not os.path.exists(dl_directory):
            os.makedirs(dl_directory)

        # Call the GitHub API
        r = requests.get(api_url, auth=nf_core.utils.github_api_auto_auth())
        if r.status_code != 200:
            raise SystemError("Could not fetch {} file: {}\n {}".format(self.name, r.status_code, api_url))
        result = r.json()
        file_contents = base64.b64decode(result["content"])

        # Write the file contents
        with open(dl_filename, "wb") as fh:
            fh.write(file_contents)


class PipelineModules(object):
    def __init__(self):
        """
        Initialise the PipelineModules object
        """
        self.modules_repo = ModulesRepo()
        self.pipeline_dir = None
        self.pipeline_module_names = []

    def list_modules(self, print_json=False):
        """
        Get available module names from GitHub tree for repo
        and print as list to stdout
        """

        # Initialise rich table
        table = rich.table.Table()
        table.add_column("Module Name")
        modules = []

        # No pipeline given - show all remote
        if self.pipeline_dir is None:
            log.info(f"Modules available from {self.modules_repo.name} ({self.modules_repo.branch}):\n")

            # Get the list of available modules
            self.modules_repo.get_modules_file_tree()
            modules = self.modules_repo.modules_avail_module_names
            # Nothing found
            if len(modules) == 0:
                log.info(f"No available modules found in {self.modules_repo.name} ({self.modules_repo.branch})")
                return ""

        # We have a pipeline - list what's installed
        else:
            log.info(f"Modules installed in '{self.pipeline_dir}':\n")

            # Check whether pipelines is valid
            try:
                self.has_valid_pipeline()
            except UserWarning as e:
                log.error(e)
                return ""
            # Get installed modules
            self.get_pipeline_modules()
            modules = self.pipeline_module_names
            # Nothing found
            if len(modules) == 0:
                log.info(f"No nf-core modules found in '{self.pipeline_dir}'")
                return ""

        for mod in sorted(modules):
            table.add_row(mod)
        if print_json:
            return json.dumps(modules, sort_keys=True, indent=4)
        return table

    def install(self, module=None):

        # Check whether pipelines is valid
        self.has_valid_pipeline()

        # Get the available modules
        self.modules_repo.get_modules_file_tree()

        if module is None:
            module = questionary.autocomplete(
                "Tool name:",
                choices=self.modules_repo.modules_avail_module_names,
                style=nf_core.utils.nfcore_question_style,
            ).ask()

        log.info("Installing {}".format(module))

        # Check that the supplied name is an available module
        if module not in self.modules_repo.modules_avail_module_names:
            log.error("Module '{}' not found in list of available modules.".format(module))
            log.info("Use the command 'nf-core modules list' to view available software")
            return False
        log.debug("Installing module '{}' at modules hash {}".format(module, self.modules_repo.modules_current_hash))

        # Check that we don't already have a folder for this module
        module_dir = os.path.join(self.pipeline_dir, "modules", "nf-core", "software", module)
        if os.path.exists(module_dir):
            log.error("Module directory already exists: {}".format(module_dir))
            # TODO: uncomment next line once update is implemented
            # log.info("To update an existing module, use the commands 'nf-core update'")
            return False

        # Download module files
        files = self.modules_repo.get_module_file_urls(module)
        log.debug("Fetching module files:\n - {}".format("\n - ".join(files.keys())))
        for filename, api_url in files.items():
            dl_filename = os.path.join(self.pipeline_dir, "modules", "nf-core", filename)
            self.modules_repo.download_gh_file(dl_filename, api_url)
        log.info("Downloaded {} files to {}".format(len(files), module_dir))

    def update(self, module, force=False):
        log.error("This command is not yet implemented")
        pass

    def remove(self, module):
        """
        Remove an already installed module
        This command only works for modules that are installed from 'nf-core/modules'
        """

        # Check whether pipelines is valid
        self.has_valid_pipeline()

        # Get the installed modules
        self.get_pipeline_modules()

        if module is None:
            if len(self.pipeline_module_names) == 0:
                log.error("No installed modules found in pipeline")
                return False
            module = questionary.autocomplete(
                "Tool name:", choices=self.pipeline_module_names, style=nf_core.utils.nfcore_question_style
            ).ask()

        # Get the module directory
        module_dir = os.path.join(self.pipeline_dir, "modules", "nf-core", "software", module)

        # Verify that the module is actually installed
        if not os.path.exists(module_dir):
            log.error("Module directory is not installed: {}".format(module_dir))
            log.info("The module you want to remove does not seem to be installed")
            return False

        log.info("Removing {}".format(module))

        # Remove the module
        try:
            shutil.rmtree(module_dir)
            # Try cleaning up empty parent if tool/subtool and tool/ is empty
            if module.count("/") > 0:
                parent_dir = os.path.dirname(module_dir)
                try:
                    os.rmdir(parent_dir)
                except OSError:
                    log.debug(f"Parent directory not empty: '{parent_dir}'")
                else:
                    log.debug(f"Deleted orphan tool directory: '{parent_dir}'")
            log.info("Successfully removed {} module".format(module))
            return True
        except OSError as e:
            log.error("Could not remove module: {}".format(e))
            return False

    def get_pipeline_modules(self):
        """ Get list of modules installed in the current pipeline """
        self.pipeline_module_names = []
        module_mains = glob.glob(f"{self.pipeline_dir}/modules/nf-core/software/**/main.nf", recursive=True)
        for mod in module_mains:
            self.pipeline_module_names.append(
                os.path.dirname(os.path.relpath(mod, f"{self.pipeline_dir}/modules/nf-core/software/"))
            )

    def has_valid_pipeline(self):
        """Check that we were given a pipeline"""
        if self.pipeline_dir is None or not os.path.exists(self.pipeline_dir):
            log.error("Could not find pipeline: {}".format(self.pipeline_dir))
            return False
        main_nf = os.path.join(self.pipeline_dir, "main.nf")
        nf_config = os.path.join(self.pipeline_dir, "nextflow.config")
        if not os.path.exists(main_nf) and not os.path.exists(nf_config):
            raise UserWarning(f"Could not find a 'main.nf' or 'nextflow.config' file in '{self.pipeline_dir}'")



#!/usr/bin/env python

import logging
import os
import requests
import yaml
import nf_core.utils

from nf_core.utils import anaconda_package

# Set up local caching for requests to speed up remote queries
nf_core.utils.setup_requests_cachedir()

# Don't pick up debug logs from the requests package
logging.getLogger("requests").setLevel(logging.WARNING)
logging.getLogger("urllib3").setLevel(logging.WARNING)

log = logging.getLogger(__name__)


def conda_env_yaml(self):
    """Checks that the conda environment file is valid.

    .. note:: This test is ignored if there is not an ``environment.yml``
              file present in the pipeline root directory.

    DSL1 nf-core pipelines use a single Conda environment to manage all software
    dependencies for a workflow. This can be used directly with ``-profile conda``
    and is also used in the ``Dockerfile`` to build a docker image.

    This test checks the conda ``environment.yml`` file to ensure that it follows nf-core guidelines.
    Each dependency is checked using the `Anaconda API service <https://api.anaconda.org/docs>`_.
    Dependency sublists are ignored with the exception of ``- pip``: these packages are also checked
    for pinned version numbers and checked using the `PyPI JSON API <https://wiki.python.org/moin/PyPIJSON>`_.

    Specifically, this lint test makes sure that:

    * The environment ``name`` must match the pipeline name and version

        * The pipeline name is defined in the config variable ``manifest.name``
        * Replace the slash with a hyphen as environment names shouldn't contain that character
        * Example: For ``nf-core/test`` version 1.4, the conda environment name should be ``nf-core-test-1.4``

    * All package dependencies have a specific version number pinned

        .. warning:: Remember that Conda package versions should be pinned with one equals sign (``toolname=1.1``),
                 but pip uses two (``toolname==1.2``)

    * That package versions can be found and are the latest available

        * Test will go through all conda channels listed in the file, or check PyPI if ``pip``
        * Conda dependencies with pinned channels (eg. ``conda-forge::openjdk``) are ok too
        * In addition to the package name, the pinned version is checked
        * If a newer version is available, a warning will be reported
    """
    passed = []
    warned = []
    failed = []
    fixed = []
    could_fix = False

    env_path = os.path.join(self.wf_path, "environment.yml")
    if env_path not in self.files:
        return {"ignored": ["No `environment.yml` file found - skipping conda_env_yaml test"]}

    with open(env_path, "r") as fh:
        raw_environment_yml = fh.read()

    # Check that the environment name matches the pipeline name
    pipeline_version = self.nf_config.get("manifest.version", "").strip(" '\"")
    expected_env_name = "nf-core-{}-{}".format(self.pipeline_name.lower(), pipeline_version)
    if self.conda_config["name"] != expected_env_name:
        if "conda_env_yaml" in self.fix:
            passed.append("Conda environment name was correct ({})".format(expected_env_name))
            fixed.append(
                "Fixed Conda environment name: '{}' to '{}'".format(self.conda_config["name"], expected_env_name)
            )
            raw_environment_yml = raw_environment_yml.replace(self.conda_config["name"], expected_env_name)
        else:
            failed.append(
                "Conda environment name is incorrect ({}, should be {})".format(
                    self.conda_config["name"], expected_env_name
                )
            )
            could_fix = True
    else:
        passed.append("Conda environment name was correct ({})".format(expected_env_name))

    # Check conda dependency list
    conda_deps = self.conda_config.get("dependencies", [])
    if len(conda_deps) > 0:
        conda_progress = self.progress_bar.add_task(
            "Checking Conda packages", total=len(conda_deps), test_name=conda_deps[0]
        )
    for idx, dep in enumerate(conda_deps):
        self.progress_bar.update(conda_progress, advance=1, test_name=dep)
        if isinstance(dep, str):
            # Check that each dependency has a version number
            try:
                assert dep.count("=") in [1, 2]
            except AssertionError:
                failed.append("Conda dep did not have pinned version number: `{}`".format(dep))
            else:
                passed.append("Conda dep had pinned version number: `{}`".format(dep))

                try:
                    depname, depver = dep.split("=")[:2]
                    self.conda_package_info[dep] = anaconda_package(
                        dep, dep_channels=self.conda_config.get("channels", [])
                    )
                except LookupError as e:
                    warned.append(e)
                except ValueError as e:
                    failed.append(e)
                else:
                    # Check that required version is available at all
                    if depver not in self.conda_package_info[dep].get("versions"):
                        failed.append("Conda dep had unknown version: {}".format(dep))
                        continue  # No need to test for latest version, continue linting
                    # Check version is latest available
                    last_ver = self.conda_package_info[dep].get("latest_version")
                    if last_ver is not None and last_ver != depver:
                        if "conda_env_yaml" in self.fix:
                            passed.append("Conda package is the latest available: `{}`".format(dep))
                            fixed.append("Conda package updated: '{}' to '{}'".format(dep, last_ver))
                            raw_environment_yml = raw_environment_yml.replace(dep, f"{depname}={last_ver}")
                        else:
                            warned.append("Conda dep outdated: `{}`, `{}` available".format(dep, last_ver))
                            could_fix = True
                    else:
                        passed.append("Conda package is the latest available: `{}`".format(dep))

        elif isinstance(dep, dict):
            pip_deps = dep.get("pip", [])
            if len(pip_deps) > 0:
                pip_progress = self.progress_bar.add_task(
                    "Checking PyPI packages", total=len(pip_deps), test_name=pip_deps[0]
                )
            for pip_idx, pip_dep in enumerate(pip_deps):
                self.progress_bar.update(pip_progress, advance=1, test_name=pip_dep)
                # Check that each pip dependency has a version number
                try:
                    assert pip_dep.count("=") == 2
                except AssertionError:
                    failed.append("Pip dependency did not have pinned version number: {}".format(pip_dep))
                else:
                    passed.append("Pip dependency had pinned version number: {}".format(pip_dep))

                    try:
                        pip_depname, pip_depver = pip_dep.split("==", 1)
                        self.conda_package_info[pip_dep] = nf_core.utils.pip_package(pip_dep)
                    except LookupError as e:
                        warned.append(e)
                    except ValueError as e:
                        failed.append(e)
                    else:
                        # Check, if PyPI package version is available at all
                        if pip_depver not in self.conda_package_info[pip_dep].get("releases").keys():
                            failed.append("PyPI package had an unknown version: {}".format(pip_depver))
                            continue  # No need to test latest version, if not available
                        pip_last_ver = self.conda_package_info[pip_dep].get("info").get("version")
                        if pip_last_ver is not None and pip_last_ver != pip_depver:
                            if "conda_env_yaml" in self.fix:
                                passed.append("PyPI package is latest available: {}".format(pip_depver))
                                fixed.append("PyPI package updated: '{}' to '{}'".format(pip_depname, pip_last_ver))
                                raw_environment_yml = raw_environment_yml.replace(pip_depver, pip_last_ver)
                            else:
                                warned.append(
                                    "PyPI package is not latest available: {}, {} available".format(
                                        pip_depver, pip_last_ver
                                    )
                                )
                                could_fix = True
                        else:
                            passed.append("PyPI package is latest available: {}".format(pip_depver))
            self.progress_bar.update(pip_progress, visible=False)
    self.progress_bar.update(conda_progress, visible=False)

    # NB: It would be a lot easier to just do a yaml.dump on the dictionary we have,
    # but this discards all formatting and comments which is a pain.
    if "conda_env_yaml" in self.fix and len(fixed) > 0:
        with open(env_path, "w") as fh:
            fh.write(raw_environment_yml)

    return {"passed": passed, "warned": warned, "failed": failed, "fixed": fixed, "could_fix": could_fix}



#!/usr/bin/env python

import logging
import os
import nf_core

log = logging.getLogger(__name__)


def conda_dockerfile(self):
    """Checks the Dockerfile for use with Conda environments

    .. note:: This test only runs if there is both an ``environment.yml``
              and ``Dockerfile`` present in the pipeline root directory.

    If a workflow has a conda ``environment.yml`` file, the ``Dockerfile`` should use this
    to create the docker image. These files are typically very short, just creating the conda
    environment inside the container.

    This linting test checks for the following:

    * All of the following lines are present in the file (where ``PIPELINE`` is your pipeline name):

        .. code-block:: Dockerfile

           FROM nfcore/base:VERSION
           COPY environment.yml /
           RUN conda env create --quiet -f /environment.yml && conda clean -a
           RUN conda env export --name PIPELINE > PIPELINE.yml
           ENV PATH /opt/conda/envs/PIPELINE/bin:$PATH

    * That the ``FROM nfcore/base:VERSION`` is tagged to the most recent release of nf-core/tools

        * The linting tool compares the tag against the currently installed version of tools.
        * This line is not checked if running a development version of nf-core/tools.

    .. tip:: Additional lines and different metadata can be added to the ``Dockerfile``
                 without causing this lint test to fail.
    """

    # Check if we have both a conda and dockerfile
    if self._fp("environment.yml") not in self.files or self._fp("Dockerfile") not in self.files:
        return {"ignored": ["No `environment.yml` / `Dockerfile` file found - skipping conda_dockerfile test"]}

    expected_strings = [
        "COPY environment.yml /",
        "RUN conda env create --quiet -f /environment.yml && conda clean -a",
        "RUN conda env export --name {} > {}.yml".format(self.conda_config["name"], self.conda_config["name"]),
        "ENV PATH /opt/conda/envs/{}/bin:$PATH".format(self.conda_config["name"]),
    ]

    if "dev" not in nf_core.__version__:
        expected_strings.append("FROM nfcore/base:{}".format(nf_core.__version__))

    with open(os.path.join(self.wf_path, "Dockerfile"), "r") as fh:
        dockerfile_contents = fh.read().splitlines()

    difference = set(expected_strings) - set(dockerfile_contents)
    if not difference:
        return {"passed": ["Found all expected strings in Dockerfile file"]}
    else:
        return {"failed": ["Could not find Dockerfile file string: `{}`".format(missing) for missing in difference]}



#!/usr/bin/env python

import logging
import os
import io
import fnmatch

log = logging.getLogger(__name__)


def merge_markers(self):
    """Check for remaining merge markers.

    This test looks for remaining merge markers in the code, e.g.:
    >>>>>>> or <<<<<<<


    """
    passed = []
    failed = []

    ignore = [".git"]
    if os.path.isfile(os.path.join(self.wf_path, ".gitignore")):
        with io.open(os.path.join(self.wf_path, ".gitignore"), "rt", encoding="latin1") as fh:
            for l in fh:
                ignore.append(os.path.basename(l.strip().rstrip("/")))
    for root, dirs, files in os.walk(self.wf_path, topdown=True):
        # Ignore files
        for i_base in ignore:
            i = os.path.join(root, i_base)
            dirs[:] = [d for d in dirs if not fnmatch.fnmatch(os.path.join(root, d), i)]
            files[:] = [f for f in files if not fnmatch.fnmatch(os.path.join(root, f), i)]
        for fname in files:
            try:
                with io.open(os.path.join(root, fname), "rt", encoding="latin1") as fh:
                    for l in fh:
                        if ">>>>>>>" in l:
                            failed.append(f"Merge marker '>>>>>>>' in `{os.path.join(root, fname)}`: {l}")
                        if "<<<<<<<" in l:
                            failed.append(f"Merge marker '<<<<<<<' in `{os.path.join(root, fname)}`: {l}")
                            print(root)
            except FileNotFoundError:
                log.debug(f"Could not open file {os.path.join(root, fname)} in merge_markers lint test")
    if len(failed) == 0:
        passed.append("No merge markers found in pipeline files")
    return {"passed": passed, "failed": failed}



#!/usr/bin/env python
"""Bumps the version number in all appropriate files for
a nf-core pipeline.
"""

import click
import logging
import os
import re
import rich.console
import sys
import nf_core.utils

log = logging.getLogger(__name__)
stderr = rich.console.Console(file=sys.stderr, force_terminal=nf_core.utils.rich_force_colors())


def bump_pipeline_version(pipeline_obj, new_version):
    """Bumps a pipeline version number.

    Args:
        pipeline_obj (nf_core.utils.Pipeline): A `Pipeline` object that holds information
            about the pipeline contents and build files.
        new_version (str): The new version tag for the pipeline. Semantic versioning only.
    """

    # Collect the old and new version numbers
    current_version = pipeline_obj.nf_config.get("manifest.version", "").strip(" '\"")
    if new_version.startswith("v"):
        log.warning("Stripping leading 'v' from new version number")
        new_version = new_version[1:]
    if not current_version:
        log.error("Could not find config variable 'manifest.version'")
        sys.exit(1)
    log.info("Changing version number from '{}' to '{}'".format(current_version, new_version))

    # nextflow.config - workflow manifest version
    # nextflow.config - process container manifest version
    docker_tag = "dev"
    if new_version.replace(".", "").isdigit():
        docker_tag = new_version
    else:
        log.info("New version contains letters. Setting docker tag to 'dev'")

    update_file_version(
        "nextflow.config",
        pipeline_obj,
        [
            (
                r"version\s*=\s*[\'\"]?{}[\'\"]?".format(current_version.replace(".", r"\.")),
                "version = '{}'".format(new_version),
            ),
            (
                r"container\s*=\s*[\'\"]nfcore/{}:(?:{}|dev)[\'\"]".format(
                    pipeline_obj.pipeline_name.lower(), current_version.replace(".", r"\.")
                ),
                "container = 'nfcore/{}:{}'".format(pipeline_obj.pipeline_name.lower(), docker_tag),
            ),
        ],
    )

    # .github/workflows/ci.yml - docker build image tag
    # .github/workflows/ci.yml - docker tag image
    update_file_version(
        os.path.join(".github", "workflows", "ci.yml"),
        pipeline_obj,
        [
            (
                r"docker build --no-cache . -t nfcore/{name}:(?:{tag}|dev)".format(
                    name=pipeline_obj.pipeline_name.lower(), tag=current_version.replace(".", r"\.")
                ),
                "docker build --no-cache . -t nfcore/{name}:{tag}".format(
                    name=pipeline_obj.pipeline_name.lower(), tag=docker_tag
                ),
            ),
            (
                r"docker tag nfcore/{name}:dev nfcore/{name}:(?:{tag}|dev)".format(
                    name=pipeline_obj.pipeline_name.lower(), tag=current_version.replace(".", r"\.")
                ),
                "docker tag nfcore/{name}:dev nfcore/{name}:{tag}".format(
                    name=pipeline_obj.pipeline_name.lower(), tag=docker_tag
                ),
            ),
        ],
    )

    # environment.yml - environment name
    update_file_version(
        "environment.yml",
        pipeline_obj,
        [
            (
                r"name: nf-core-{}-{}".format(pipeline_obj.pipeline_name.lower(), current_version.replace(".", r"\.")),
                "name: nf-core-{}-{}".format(pipeline_obj.pipeline_name.lower(), new_version),
            )
        ],
    )

    # Dockerfile - ENV PATH and RUN conda env create
    update_file_version(
        "Dockerfile",
        pipeline_obj,
        [
            (
                r"nf-core-{}-{}".format(pipeline_obj.pipeline_name.lower(), current_version.replace(".", r"\.")),
                "nf-core-{}-{}".format(pipeline_obj.pipeline_name.lower(), new_version),
            )
        ],
    )


def bump_nextflow_version(pipeline_obj, new_version):
    """Bumps the required Nextflow version number of a pipeline.

    Args:
        pipeline_obj (nf_core.utils.Pipeline): A `Pipeline` object that holds information
            about the pipeline contents and build files.
        new_version (str): The new version tag for the required Nextflow version.
    """

    # Collect the old and new version numbers - strip leading non-numeric characters (>=)
    current_version = pipeline_obj.nf_config.get("manifest.nextflowVersion", "").strip(" '\"")
    current_version = re.sub(r"^[^0-9\.]*", "", current_version)
    new_version = re.sub(r"^[^0-9\.]*", "", new_version)
    if not current_version:
        log.error("Could not find config variable 'manifest.nextflowVersion'")
        sys.exit(1)
    log.info("Changing Nextlow version number from '{}' to '{}'".format(current_version, new_version))

    # nextflow.config - manifest minimum nextflowVersion
    update_file_version(
        "nextflow.config",
        pipeline_obj,
        [
            (
                r"nextflowVersion\s*=\s*[\'\"]?>={}[\'\"]?".format(current_version.replace(".", r"\.")),
                "nextflowVersion = '>={}'".format(new_version),
            )
        ],
    )

    # .github/workflows/ci.yml - Nextflow version matrix
    update_file_version(
        os.path.join(".github", "workflows", "ci.yml"),
        pipeline_obj,
        [
            (
                r"nxf_ver: \[[\'\"]?{}[\'\"]?, '21.03.0-edge'\]".format(current_version.replace(".", r"\.")),
                "nxf_ver: ['{}', '21.03.0-edge']".format(new_version),
            )
        ],
    )

    # README.md - Nextflow version badge
    update_file_version(
        "README.md",
        pipeline_obj,
        [
            (
                r"nextflow-%E2%89%A5{}-brightgreen.svg".format(current_version.replace(".", r"\.")),
                "nextflow-%E2%89%A5{}-brightgreen.svg".format(new_version),
            )
        ],
    )


def update_file_version(filename, pipeline_obj, patterns):
    """Updates the version number in a requested file.

    Args:
        filename (str): File to scan.
        pipeline_obj (nf_core.lint.PipelineLint): A PipelineLint object that holds information
            about the pipeline contents and build files.
        pattern (str): Regex pattern to apply.
        newstr (str): The replaced string.

    Raises:
        ValueError, if the version number cannot be found.
    """
    # Load the file
    fn = pipeline_obj._fp(filename)
    content = ""
    try:
        with open(fn, "r") as fh:
            content = fh.read()
    except FileNotFoundError:
        log.warning("File not found: '{}'".format(fn))
        return

    replacements = []
    for pattern in patterns:

        # Check that we have a match
        matches_pattern = re.findall("^.*{}.*$".format(pattern[0]), content, re.MULTILINE)
        if len(matches_pattern) == 0:
            log.error("Could not find version number in {}: '{}'".format(filename, pattern))
            continue

        # Replace the match
        content = re.sub(pattern[0], pattern[1], content)
        matches_newstr = re.findall("^.*{}.*$".format(pattern[1]), content, re.MULTILINE)

        # Save for logging
        replacements.append((matches_pattern, matches_newstr))

    log.info("Updated version in '{}'".format(filename))
    for replacement in replacements:
        for idx, matched in enumerate(replacement[0]):
            stderr.print("          [red] - {}".format(matched.strip()), highlight=False)
            stderr.print("          [green] + {}".format(replacement[1][idx].strip()), highlight=False)
    stderr.print("\n")

    with open(fn, "w") as fh:
        fh.write(content)



#!/usr/bin/env python

import logging
import yaml
import json
import jsonschema
import os
import glob
import requests


def actions_schema_validation(self):
    """Checks that the GitHub Action workflow yml/yaml files adhere to the correct schema

    nf-core pipelines use GitHub actions workflows to run CI tests, check formatting and also linting, among others.
    These workflows are defined by ``yml``scripts in ``.github/workflows/``. This lint test verifies that these scripts are valid
    by comparing them against the JSON schema for GitHub workflows <https://json.schemastore.org/github-workflow>

    To pass this test, make sure that all your workflows contain the required properties ``on`` and ``jobs``and that
    all other properties are of the correct type, as specified in the schema (link above).
    """
    passed = []
    failed = []

    # Only show error messages from schema
    logging.getLogger("nf_core.schema").setLevel(logging.ERROR)

    # Get all workflow files
    action_workflows = glob.glob(os.path.join(self.wf_path, ".github/workflows/*.y*ml"))

    # Load the GitHub workflow schema
    r = requests.get("https://json.schemastore.org/github-workflow", allow_redirects=True)
    schema = r.json()

    # Validate all workflows against the schema
    for wf_path in action_workflows:
        wf = os.path.basename(wf_path)

        # load workflow
        try:
            with open(wf_path, "r") as fh:
                wf_json = yaml.safe_load(fh)
        except Exception as e:
            failed.append("Could not parse yaml file: {}, {}".format(wf, e))
            continue

        # yaml parses 'on' as True --> try to fix it before schema validation
        try:
            wf_json["on"] = wf_json.pop(True)
        except Exception as e:
            failed.append("Missing 'on' keyword in {}.format(wf)")

        # Validate the workflow
        try:
            jsonschema.validate(wf_json, schema)
            passed.append("Workflow validation passed: {}".format(wf))
        except Exception as e:
            failed.append("Workflow validation failed for {}: {}".format(wf, e))

    return {"passed": passed, "failed": failed}



#!/usr/bin/env python

import os
import re


def readme(self):
    """Repository ``README.md`` tests

    The ``README.md`` files for a project are very important and must meet some requirements:

    * Nextflow badge

      * If no Nextflow badge is found, a warning is given
      * If a badge is found but the version doesn't match the minimum version in the config file, the test fails
      * Example badge code:

        .. code-block:: md

           [![Nextflow](https://img.shields.io/badge/nextflow-%E2%89%A50.27.6-brightgreen.svg)](https://www.nextflow.io/)

    * Bioconda badge

      * If your pipeline contains a file called ``environment.yml`` in the root directory, a bioconda badge is required
      * Required badge code:

        .. code-block:: md

           [![install with bioconda](https://img.shields.io/badge/install%20with-bioconda-brightgreen.svg)](https://bioconda.github.io/)

    .. note:: These badges are a markdown image ``![alt-text](<image URL>)`` *inside* a markdown link ``[markdown image](<link URL>)``, so a bit fiddly to write.
    """
    passed = []
    warned = []
    failed = []

    with open(os.path.join(self.wf_path, "README.md"), "r") as fh:
        content = fh.read()

    # Check that there is a readme badge showing the minimum required version of Nextflow
    # and that it has the correct version
    nf_badge_re = r"\[!\[Nextflow\]\(https://img\.shields\.io/badge/nextflow-%E2%89%A5([\d\.]+)-brightgreen\.svg\)\]\(https://www\.nextflow\.io/\)"
    match = re.search(nf_badge_re, content)
    if match:
        nf_badge_version = match.group(1).strip("'\"")
        try:
            assert nf_badge_version == self.minNextflowVersion
        except (AssertionError, KeyError):
            failed.append(
                "README Nextflow minimum version badge does not match config. Badge: `{}`, Config: `{}`".format(
                    nf_badge_version, self.minNextflowVersion
                )
            )
        else:
            passed.append(
                "README Nextflow minimum version badge matched config. Badge: `{}`, Config: `{}`".format(
                    nf_badge_version, self.minNextflowVersion
                )
            )
    else:
        warned.append("README did not have a Nextflow minimum version badge.")

    # Check that we have a bioconda badge if we have a bioconda environment file
    if os.path.join(self.wf_path, "environment.yml") in self.files:
        bioconda_badge = "[![install with bioconda](https://img.shields.io/badge/install%20with-bioconda-brightgreen.svg)](https://bioconda.github.io/)"
        if bioconda_badge in content:
            passed.append("README had a bioconda badge")
        else:
            warned.append("Found a bioconda environment.yml file but no badge in the README")

    return {"passed": passed, "warned": warned, "failed": failed}



#!/usr/bin/env python

import io
import mimetypes
import re


def template_strings(self):
    """Check for template placeholders.

    The ``nf-core create`` pipeline template uses
    `Jinja <https://jinja.palletsprojects.com/en/2.11.x/>`_ behind the scenes.

    This lint test fails if any Jinja template variables such as
    ``{{ pipeline_name }}`` are found in your pipeline code.

    Finding a placeholder like this means that something was probably copied and pasted
    from the template without being properly rendered for your pipeline.

    This test ignores any double-brackets prefixed with a dollar sign, such as
    ``${{ secrets.AWS_ACCESS_KEY_ID }}`` as these placeholders are used in GitHub Actions workflows.
    """
    passed = []
    failed = []

    # Loop through files, searching for string
    num_matches = 0
    for fn in self.files:

        # Skip binary files
        binary_ftypes = ["image", "application/java-archive"]
        (ftype, encoding) = mimetypes.guess_type(fn)
        if encoding is not None or (ftype is not None and any([ftype.startswith(ft) for ft in binary_ftypes])):
            continue

        with io.open(fn, "r", encoding="latin1") as fh:
            lnum = 0
            for l in fh:
                lnum += 1
                cc_matches = re.findall(r"[^$]{{[^:}]*}}", l)
                if len(cc_matches) > 0:
                    for cc_match in cc_matches:
                        failed.append("Found a Jinja template string in `{}` L{}: {}".format(fn, lnum, cc_match))
                        num_matches += 1
    if num_matches == 0:
        passed.append("Did not find any Jinja template strings ({} files)".format(len(self.files)))

    return {"passed": passed, "failed": failed}



#!/usr/bin/env python


def pipeline_name_conventions(self):
    """Checks that the pipeline name adheres to nf-core conventions.

    In order to ensure consistent naming, pipeline names should contain only lower case, alphanumeric characters.
    Otherwise a warning is displayed.

    .. warning::
        DockerHub is very picky about image names and doesn't even allow hyphens (we are ``nfcore``).
        This is a large part of why we set this rule.
    """
    passed = []
    warned = []
    failed = []

    if self.pipeline_name.islower() and self.pipeline_name.isalnum():
        passed.append("Name adheres to nf-core convention")
    if not self.pipeline_name.islower():
        warned.append("Naming does not adhere to nf-core conventions: Contains uppercase letters")
    if not self.pipeline_name.isalnum():
        warned.append("Naming does not adhere to nf-core conventions: Contains non alphanumeric characters")

    return {"passed": passed, "warned": warned, "failed": failed}



#!/usr/bin/env python
"""
The ModulesTestYmlBuilder class handles automatic generation of the modules test.yml file
along with running the tests and creating md5 sums
"""

from __future__ import print_function
from rich.syntax import Syntax

import errno
import hashlib
import logging
import os
import questionary
import re
import rich
import shlex
import subprocess
import tempfile
import yaml
import operator

import nf_core.utils
import nf_core.modules.pipeline_modules


log = logging.getLogger(__name__)


class ModulesTestYmlBuilder(object):
    def __init__(
        self,
        module_name=None,
        run_tests=False,
        test_yml_output_path=None,
        force_overwrite=False,
        no_prompts=False,
    ):
        self.module_name = module_name
        self.run_tests = run_tests
        self.test_yml_output_path = test_yml_output_path
        self.force_overwrite = force_overwrite
        self.no_prompts = no_prompts
        self.module_dir = None
        self.module_test_main = None
        self.entry_points = []
        self.tests = []

    def run(self):
        """ Run build steps """
        if not self.no_prompts:
            log.info(
                "[yellow]Press enter to use default values [cyan bold](shown in brackets) [yellow]or type your own responses"
            )
        self.check_inputs()
        self.scrape_workflow_entry_points()
        self.build_all_tests()
        self.print_test_yml()

    def check_inputs(self):
        """ Do more complex checks about supplied flags. """

        # Get the tool name if not specified
        if self.module_name is None:
            modules_repo = nf_core.modules.pipeline_modules.ModulesRepo()
            modules_repo.get_modules_file_tree()
            self.module_name = questionary.autocomplete(
                "Tool name:",
                choices=modules_repo.modules_avail_module_names,
                style=nf_core.utils.nfcore_question_style,
            ).ask()
        self.module_dir = os.path.join("software", *self.module_name.split("/"))
        self.module_test_main = os.path.join("tests", "software", *self.module_name.split("/"), "main.nf")

        # First, sanity check that the module directory exists
        if not os.path.isdir(self.module_dir):
            raise UserWarning(f"Cannot find directory '{self.module_dir}'. Should be TOOL/SUBTOOL or TOOL")
        if not os.path.exists(self.module_test_main):
            raise UserWarning(f"Cannot find module test workflow '{self.module_test_main}'")

        # Check that we're running tests if no prompts
        if not self.run_tests and self.no_prompts:
            log.debug("Setting run_tests to True as running without prompts")
            self.run_tests = True

        # Get the output YAML file / check it does not already exist
        while self.test_yml_output_path is None:
            default_val = f"tests/software/{self.module_name}/test.yml"
            if self.no_prompts:
                self.test_yml_output_path = default_val
            else:
                self.test_yml_output_path = rich.prompt.Prompt.ask(
                    "[violet]Test YAML output path[/] (- for stdout)", default=default_val
                ).strip()
                if self.test_yml_output_path == "":
                    self.test_yml_output_path = None
                # Check that the output YAML file does not already exist
                if (
                    self.test_yml_output_path is not None
                    and self.test_yml_output_path != "-"
                    and os.path.exists(self.test_yml_output_path)
                    and not self.force_overwrite
                ):
                    if rich.prompt.Confirm.ask(
                        f"[red]File exists! [green]'{self.test_yml_output_path}' [violet]Overwrite?"
                    ):
                        self.force_overwrite = True
                    else:
                        self.test_yml_output_path = None
        if os.path.exists(self.test_yml_output_path) and not self.force_overwrite:
            raise UserWarning(
                f"Test YAML file already exists! '{self.test_yml_output_path}'. Use '--force' to overwrite."
            )

    def scrape_workflow_entry_points(self):
        """ Find the test workflow entry points from main.nf """
        log.info(f"Looking for test workflow entry points: '{self.module_test_main}'")
        with open(self.module_test_main, "r") as fh:
            for line in fh:
                match = re.match(r"workflow\s+(\S+)\s+{", line)
                if match:
                    self.entry_points.append(match.group(1))
        if len(self.entry_points) == 0:
            raise UserWarning("No workflow entry points found in 'self.module_test_main'")

    def build_all_tests(self):
        """
        Go over each entry point and build structure
        """
        for entry_point in self.entry_points:
            ep_test = self.build_single_test(entry_point)
            if ep_test:
                self.tests.append(ep_test)

    def build_single_test(self, entry_point):
        """Given the supplied cli flags, prompt for any that are missing.

        Returns: Test command
        """
        ep_test = {
            "name": "",
            "command": "",
            "tags": [],
            "files": [],
        }

        # Print nice divider line
        console = rich.console.Console()
        console.print("[black]" + "─" * console.width)

        log.info(f"Building test meta for entry point '{entry_point}'")

        while ep_test["name"] == "":
            default_val = f"{self.module_name.replace('/', ' ')} {entry_point}"
            if self.no_prompts:
                ep_test["name"] = default_val
            else:
                ep_test["name"] = rich.prompt.Prompt.ask("[violet]Test name", default=default_val).strip()

        while ep_test["command"] == "":
            default_val = (
                f"nextflow run tests/software/{self.module_name} -entry {entry_point} -c tests/config/nextflow.config"
            )
            if self.no_prompts:
                ep_test["command"] = default_val
            else:
                ep_test["command"] = rich.prompt.Prompt.ask("[violet]Test command", default=default_val).strip()

        while len(ep_test["tags"]) == 0:
            mod_name_parts = self.module_name.split("/")
            tag_defaults = []
            for idx in range(0, len(mod_name_parts)):
                tag_defaults.append("_".join(mod_name_parts[: idx + 1]))
            tag_defaults.append(entry_point.replace("test_", ""))
            # Remove duplicates
            tag_defaults = list(set(tag_defaults))
            if self.no_prompts:
                ep_test["tags"] = tag_defaults
            else:
                while len(ep_test["tags"]) == 0:
                    prompt_tags = rich.prompt.Prompt.ask(
                        "[violet]Test tags[/] (comma separated)", default=",".join(tag_defaults)
                    ).strip()
                    ep_test["tags"] = [t.strip() for t in prompt_tags.split(",")]

        ep_test["files"] = self.get_md5_sums(entry_point, ep_test["command"])

        return ep_test

    def _md5(self, fname):
        """Generate md5 sum for file"""
        hash_md5 = hashlib.md5()
        with open(fname, "rb") as f:
            for chunk in iter(lambda: f.read(4096), b""):
                hash_md5.update(chunk)
        md5sum = hash_md5.hexdigest()
        return md5sum

    def create_test_file_dict(self, results_dir):
        """ Walk through directory and collect md5 sums """
        test_files = []
        for root, dir, file in os.walk(results_dir):
            for elem in file:
                elem = os.path.join(root, elem)
                elem_md5 = self._md5(elem)
                # Switch out the results directory path with the expected 'output' directory
                elem = elem.replace(results_dir, "output")
                test_files.append({"path": elem, "md5sum": elem_md5})

        test_files = sorted(test_files, key=operator.itemgetter("path"))

        return test_files

    def get_md5_sums(self, entry_point, command):
        """
        Recursively go through directories and subdirectories
        and generate tuples of (<file_path>, <md5sum>)
        returns: list of tuples
        """

        results_dir = None
        run_this_test = False
        while results_dir is None:
            if self.run_tests or run_this_test:
                results_dir, results_dir_repeat = self.run_tests_workflow(command)
            else:
                results_dir = rich.prompt.Prompt.ask(
                    f"[violet]Test output folder with results[/] (leave blank to run test)"
                )
                if results_dir == "":
                    results_dir = None
                    run_this_test = True
                elif not os.path.isdir(results_dir):
                    log.error(f"Directory '{results_dir}' does not exist")
                    results_dir = None

        test_files = self.create_test_file_dict(results_dir=results_dir)
        test_files_repeat = self.create_test_file_dict(results_dir=results_dir_repeat)

        # Compare both test.yml files
        for i in range(len(test_files)):
            if not test_files[i]["md5sum"] == test_files_repeat[i]["md5sum"]:
                test_files[i].pop("md5sum")
                test_files[i][
                    "contains"
                ] = "# TODO nf-core: file md5sum was variable, please replace this text with a string found in the file instead"

        if len(test_files) == 0:
            raise UserWarning(f"Could not find any test result files in '{results_dir}'")

        return test_files

    def run_tests_workflow(self, command):
        """ Given a test workflow and an entry point, run the test workflow """

        # The config expects $PROFILE and Nextflow fails if it's not set
        if os.environ.get("PROFILE") is None:
            os.environ["PROFILE"] = ""
            if self.no_prompts:
                log.info(
                    "Setting env var '$PROFILE' to an empty string as not set.\n"
                    "Tests will run with Docker by default. "
                    "To use Singularity set 'export PROFILE=singularity' in your shell before running this command."
                )
            else:
                question = {
                    "type": "list",
                    "name": "profile",
                    "message": "Choose software profile",
                    "choices": ["Docker", "Singularity", "Conda"],
                }
                answer = questionary.unsafe_prompt([question], style=nf_core.utils.nfcore_question_style)
                profile = answer["profile"].lower()
                if profile in ["singularity", "conda"]:
                    os.environ["PROFILE"] = profile
                    log.info(f"Setting env var '$PROFILE' to '{profile}'")

        tmp_dir = tempfile.mkdtemp()
        tmp_dir_repeat = tempfile.mkdtemp()
        command += f" --outdir {tmp_dir}"
        command_repeat = command + f" --outdir {tmp_dir_repeat}"

        log.info(f"Running '{self.module_name}' test with command:\n[violet]{command}")
        try:
            nfconfig_raw = subprocess.check_output(shlex.split(command))
            log.info(f"Repeating test ...")
            nfconfig_raw = subprocess.check_output(shlex.split(command_repeat))

        except OSError as e:
            if e.errno == errno.ENOENT and command.strip().startswith("nextflow "):
                raise AssertionError(
                    "It looks like Nextflow is not installed. It is required for most nf-core functions."
                )
        except subprocess.CalledProcessError as e:
            raise UserWarning(f"Error running test workflow (exit code {e.returncode})\n[red]{e.output.decode()}")
        except Exception as e:
            raise UserWarning(f"Error running test workflow: {e}")
        else:
            log.info("Test workflow finished!")
            log.debug(nfconfig_raw)

        return tmp_dir, tmp_dir_repeat

    def print_test_yml(self):
        """
        Generate the test yml file.
        """

        if self.test_yml_output_path == "-":
            console = rich.console.Console()
            yaml_str = yaml.dump(self.tests, Dumper=nf_core.utils.custom_yaml_dumper(), width=10000000)
            console.print("\n", Syntax(yaml_str, "yaml"), "\n")
            return

        try:
            log.info(f"Writing to '{self.test_yml_output_path}'")
            with open(self.test_yml_output_path, "w") as fh:
                yaml.dump(self.tests, fh, Dumper=nf_core.utils.custom_yaml_dumper(), width=10000000)
        except FileNotFoundError as e:
            raise UserWarning("Could not create test.yml file: '{}'".format(e))



#!/usr/bin/env python

import nf_core.schema


def schema_params(self):
    """Check that the schema describes all flat params in the pipeline.

    The ``nextflow_schema.json`` pipeline schema should describe every flat parameter
    returned from the ``nextflow config`` command (params that are objects or more complex structures are ignored).

    * Failure: If parameters are found in ``nextflow_schema.json`` that are not in ``nextflow_schema.json``
    * Warning: If parameters are found in ``nextflow_schema.json`` that are not in ``nextflow_schema.json``
    """
    passed = []
    warned = []
    failed = []

    # First, get the top-level config options for the pipeline
    # Schema object already created in the `schema_lint` test
    self.schema_obj.get_schema_path(self.wf_path)
    self.schema_obj.get_wf_params()
    self.schema_obj.no_prompts = True

    # Remove any schema params not found in the config
    removed_params = self.schema_obj.remove_schema_notfound_configs()

    # Add schema params found in the config but not the schema
    added_params = self.schema_obj.add_schema_found_configs()

    if len(removed_params) > 0:
        for param in removed_params:
            warned.append("Schema param `{}` not found from nextflow config".format(param))

    if len(added_params) > 0:
        for param in added_params:
            failed.append("Param `{}` from `nextflow config` not found in nextflow_schema.json".format(param))

    if len(removed_params) == 0 and len(added_params) == 0:
        passed.append("Schema matched params returned from nextflow config")

    return {"passed": passed, "warned": warned, "failed": failed}
